import logging

from airflow import settings
from airflow.models import Connection
from sqlalchemy.orm import exc

from dbnd._core.errors import DatabandRuntimeError


try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


logger = logging.getLogger(__name__)


def delete_connection(conn_id):
    session = settings.Session()
    try:
        to_delete = (
            session.query(Connection).filter(Connection.conn_id == conn_id).one()
        )
    except exc.NoResultFound:
        logger.info("Did not find a connection with `conn_id`=%s\n'", conn_id)
    except exc.MultipleResultsFound:
        raise DatabandRuntimeError(
            "Found more than one connection with "
            + "`conn_id`={conn_id}\n".format(conn_id=conn_id)
        )
    else:
        session.delete(to_delete)
        session.commit()
        logger.info("Successfully deleted `conn_id`=%s\n", conn_id)


def add_connection(
    conn_id,
    uri=None,
    conn_type=None,
    host=None,
    login=None,
    password=None,
    port=None,
    schema=None,
    extra=None,
):
    if uri:
        new_conn = Connection(conn_id=conn_id, uri=uri)
    else:
        new_conn = Connection(
            conn_id=conn_id,
            conn_type=conn_type,
            host=host,
            login=login,
            password=password,
            schema=schema,
            port=port,
        )
    if extra is not None:
        new_conn.set_extra(extra)

    session = settings.Session()
    if not (
        session.query(Connection).filter(Connection.conn_id == new_conn.conn_id).first()
    ):
        session.add(new_conn)
        session.commit()
        msg = "\n\tSuccessfully added `conn_id`={conn_id} : {uri}\n"
        msg = msg.format(
            conn_id=new_conn.conn_id,
            uri=uri
            or urlunparse(
                (
                    conn_type,
                    "{login}:{password}@{host}:{port}".format(
                        login=login or "",
                        password=password or "",
                        host=host or "",
                        port=port or "",
                    ),
                    schema or "",
                    "",
                    "",
                    "",
                )
            ),
        )
        logger.info(msg)
        logger.info("extra=%s" % str(extra))
    else:
        msg = "\n\tA connection with `conn_id`={conn_id} already exists\n"
        raise DatabandRuntimeError(msg.format(conn_id=new_conn.conn_id))


def set_connection(
    conn_id,
    uri=None,
    conn_type=None,
    host=None,
    login=None,
    password=None,
    port=None,
    schema=None,
    extra=None,
):
    delete_connection(conn_id)
    add_connection(
        conn_id=conn_id,
        uri=uri,
        conn_type=conn_type,
        host=host,
        login=login,
        password=password,
        port=port,
        schema=schema,
        extra=extra,
    )



from __future__ import absolute_import

import logging

from airflow.models import DagBag, DagModel, DagPickle, TaskInstance
from airflow.utils.db import provide_session

from dbnd._vendor import pendulum


logger = logging.getLogger(__name__)


class DbndDagModel(DagModel):
    def get_dag(self, store_serialized_dags=False):
        # DBND PATCH
        # unwrap all old logic, as we have recursion call there that makes it not easy to patch

        dag = DagBag(
            dag_folder=self.fileloc, store_serialized_dags=store_serialized_dags
        ).get_dag(self.dag_id)

        if store_serialized_dags and dag is None:
            dag = DagBag(dag_folder=self.fileloc, store_serialized_dags=False).get_dag(
                self.dag_id
            )
        if dag:
            return dag

        from airflow.www_rbac.views import dagbag

        return dagbag.get_dag(dag_id=self.dag_id)


class DbndAirflowDagBag(DagBag):
    @provide_session
    def get_dag(self, dag_id, from_file_only=True, execution_date=None, session=None):
        """
        :param dag_id:
        :param execution_date: if provided, we'll try to find specifc version of dag (using pickle)
        :param session:
        :return:
        """
        from flask import has_request_context, request, session as flask_session

        # all legacy airflow code works just with dag_id, also, there are some calls that doesn't pass through execution_date
        if has_request_context():
            execution_date = execution_date or request.args.get("execution_date")

            # trick to store last execution date used for the next flask call
            if execution_date:
                logger.debug(
                    "Execution date saved to session: %s, %s", dag_id, execution_date
                )
                flask_session["ed_" + dag_id] = execution_date
            else:
                logger.debug("Execution date from previous session: %s", dag_id)
                execution_date = flask_session.get("ed_" + dag_id)

            if execution_date and execution_date != "undefined":
                # we are going to return most "active" dag
                dttm = pendulum.parse(execution_date)
                dag = self._get_pickled_dag_from_dagrun(
                    dag_id=dag_id, execution_date=dttm, session=session
                )
                if dag:
                    return dag

        # we don't have specific dag/execution date, we are trying to get in-memory version
        dag = super(DbndAirflowDagBag, self).get_dag(
            dag_id, from_file_only=from_file_only
        )
        if dag:
            return dag

        # let try to find it latest version in DB
        latest_execution = (
            session.query(TaskInstance.execution_date)
            .filter(TaskInstance.task_id == dag_id)
            .order_by(TaskInstance.execution_date.desc())
            .first()
        )

        if latest_execution:
            return self._get_pickled_dag_from_dagrun(
                dag_id=dag_id,
                execution_date=latest_execution.execution_date,
                session=session,
            )

        return None

    @provide_session
    def _get_pickled_dag_from_dagrun(self, dag_id, execution_date, session=None):

        ti = (
            session.query(TaskInstance.task_id, TaskInstance.executor_config)
            .filter(
                TaskInstance.dag_id == dag_id,
                TaskInstance.execution_date == execution_date,
            )
            .first()
        )
        if not ti:
            logger.debug("Failed to find task instance %s %s", dag_id, execution_date)
            return None

        pickled_dag_id = ti.executor_config.get("DatabandExecutor", {}).get(
            "dag_pickle_id", None
        )
        if not pickled_dag_id:
            logger.debug(
                "No dbnd config at %s %s %s, no pickle_id",
                dag_id,
                execution_date,
                ti.task_id,
            )
            return None

        try:
            pickled_dag = (
                session.query(DagPickle)
                .filter(DagPickle.id == pickled_dag_id)
                .one_or_none()
            )
        except Exception as ex:
            logger.error("Error ocured during DAG retrieval from DB, %s", ex)
            return None

        if pickled_dag and pickled_dag.pickle:
            # we found pickled dag
            dag = pickled_dag.pickle
            dag.dag_version_execution_date = execution_date
            # let's add this dag into dags, there is a check that validates if this dag exists..
            self.dags[dag_id] = dag
            return dag
        else:
            logger.debug(
                "Failed to find pickled dag in DB for pickle_id=%s", pickled_dag_id
            )
            # failed to parse dag?
            return None



class KubernetesMetricsLogger(object):
    def __init__(self):
        pass

    def log_pod_information(self, task, pod_name, node_name):
        task.log_system_metric("pod_name", pod_name)
        task.log_system_metric("node_name", node_name)

    def log_pod_started(self, task):
        import datetime

        timestamp = datetime.datetime.utcnow().isoformat()
        task.log_system_metric("pod_started_execution", timestamp)

    def log_pod_deleted(self, task):
        import datetime

        timestamp = datetime.datetime.utcnow().isoformat()
        task.log_system_metric("pod_finished_execution", timestamp)



import logging

from types import FunctionType, ModuleType

from dbnd import task
from dbnd._core.configuration.environ_config import get_dbnd_project_config
from dbnd._core.decorator.dbnd_func_proxy import DbndFuncProxy
from dbnd._core.settings import CoreConfig
from dbnd._core.tracking.no_tracking import should_not_track
from dbnd_airflow.tracking.dbnd_spark_conf import (
    _get_databand_url,
    dbnd_wrap_spark_environment,
    get_dbnd_tracking_spark_conf_dict,
    spark_submit_with_dbnd_tracking,
)


logger = logging.getLogger(__name__)


def track_emr_add_steps_operator(operator):
    for step in operator.steps:
        args = step["HadoopJarStep"]["Args"]
        if args and "spark-submit" in args[0]:
            step["HadoopJarStep"]["Args"] = spark_submit_with_dbnd_tracking(args)


def track_databricks_submit_run_operator(operator):
    config = operator.json
    # passing env variables is only supported in new clusters
    if "new_cluster" in config:
        cluster = config["new_cluster"]
        if "spark_env_vars" not in cluster:
            cluster["spark_env_vars"] = {}
        spark_vars = cluster["spark_env_vars"]
        spark_vars["AIRFLOW_CTX_DAG_ID"] = "{{ dag.dag_id }}"
        spark_vars["AIRFLOW_CTX_EXECUTION_DATE"] = "{{ ts }}"
        spark_vars["AIRFLOW_CTX_TASK_ID"] = "{{ task.task_id }}"
        spark_vars["AIRFLOW_CTX_TRY_NUMBER"] = "{{ task_instance._try_number }}"

        databand_url = _get_databand_url()
        if databand_url:
            spark_vars["DBND__CORE__DATABAND_URL"] = databand_url


def track_data_proc_pyspark_operator(operator):
    if operator.dataproc_properties is None:
        operator.dataproc_properties = dict()
    dbnd_conf = get_dbnd_tracking_spark_conf_dict()
    operator.dataproc_properties.update(dbnd_conf)


def track_spark_submit_operator(operator):
    if operator._conf is None:
        operator._conf = dict()
    dbnd_conf = get_dbnd_tracking_spark_conf_dict()
    operator._conf.update(dbnd_conf)

    if operator._env_vars is None:
        operator._env_vars = dict()
    dbnd_env_vars = dbnd_wrap_spark_environment()
    operator._env_vars.update(dbnd_env_vars)


def is_instance_by_class_name(obj, class_name):
    for cls in obj.__class__.mro():
        if cls.__name__ == class_name:
            return True
    return False


def track_python_operator(operator):
    if isinstance(operator.python_callable, DbndFuncProxy):
        # function already has @task
        return

    operator.python_callable = task(operator.python_callable)


def _track_task(task):
    if should_not_track(task):
        return
    if is_instance_by_class_name(task, "SubDagOperator"):
        track_dag(task.subdag)
    elif is_instance_by_class_name(task, "EmrAddStepsOperator"):
        track_emr_add_steps_operator(task)
    elif is_instance_by_class_name(task, "DatabricksSubmitRunOperator"):
        track_databricks_submit_run_operator(task)
    elif is_instance_by_class_name(task, "DataProcPySparkOperator"):
        track_data_proc_pyspark_operator(task)
    elif is_instance_by_class_name(task, "SparkSubmitOperator"):
        track_spark_submit_operator(task)
    elif is_instance_by_class_name(task, "PythonOperator"):
        track_python_operator(task)


def _is_verbose():
    config = get_dbnd_project_config()
    return config.is_verbose()


def track_task(task):
    try:
        _track_task(task)
    except Exception:
        if _is_verbose():
            logger.exception("Failed to modify %s for tracking" % task.task_id)
        else:
            logger.info("Failed to modify %s for tracking" % task.task_id)


def track_dag(dag):
    """
    Modify operators in dag if necessary so we could track them.
    Supported operators: EmrAddStepsOperator, DataProcPySparkOperator, SparkSubmitOperator, PythonOperator.
    Other operators are not modified.

    More details on each operator:
    - EmrAddStepsOperator: modify args of spark-submit steps by adding dbnd variables using --conf
    - DataProcPySparkOperator: add dbnd variables to dataproc properties
    - SparkSubmitOperator: add dbnd variables to spark-submit command using --conf
    - PythonOperator: wrap python function with @task
    """
    try:
        if should_not_track(dag):
            return

        for task in dag.tasks:
            track_task(task)
    except Exception:
        logger.exception("Failed to modify %s for tracking" % dag.dag_id)



import os

from dbnd._core.settings import CoreConfig


def _filter_vars(key, bypass_dbnd, bypass_airflow, bypass_rest):
    if key.startswith("AIRFLOW_"):
        return True if bypass_airflow else False
    if key.startswith("DBND_"):
        return True if bypass_dbnd else False
    return True if bypass_rest else False


def dbnd_tracking_env(
    new_vars, bypass_dbnd=True, bypass_airflow=True, bypass_rest=False
):

    environment = {
        key: os.environ[key]
        for key in os.environ.keys()
        if _filter_vars(key, bypass_dbnd, bypass_airflow, bypass_rest)
    }
    environment.update(new_vars)
    return environment


def dbnd_wrap_spark_environment(environment=None):
    if not environment:
        environment = dict()
    environment.update({"DBND__ENABLE__SPARK_CONTEXT_ENV": "1"})
    return environment


def with_dbnd_tracking_spark_conf(user_dict):
    dbnd_dict = get_dbnd_tracking_spark_conf_dict()
    user_dict.update(dbnd_dict)
    return user_dict


def get_dbnd_tracking_spark_conf_dict(
    dag_id="{{dag.dag_id}}",
    task_id="{{task.task_id}}",
    execution_date="{{ts}}",
    try_number="{{task_instance._try_number}}",
):
    spark_conf = {
        "spark.env.AIRFLOW_CTX_DAG_ID": dag_id,
        "spark.env.AIRFLOW_CTX_EXECUTION_DATE": execution_date,
        "spark.env.AIRFLOW_CTX_TASK_ID": task_id,
        "spark.env.AIRFLOW_CTX_TRY_NUMBER": try_number,
    }

    databand_url = _get_databand_url()
    if databand_url:
        spark_conf["spark.env.DBND__CORE__DATABAND_URL"] = databand_url

    return spark_conf


def _get_databand_url():
    try:
        return CoreConfig().databand_external_url
    except Exception:
        pass


def get_dbnd_tracking_spark_conf(
    dag_id="{{dag.dag_id}}",
    task_id="{{task.task_id}}",
    execution_date="{{ts}}",
    try_number="{{task_instance._try_number}}",
):
    """ This functions returns Airflow ids  as spark configuration properties. This is used to associate spark run with airflow dag/task.
        These properties are
            spark.env.AIRFLOW_CTX_DAG_ID  -  name of the Airflow DAG to associate a run with
            spark.env.AIRFLOW_CTX_EXECUTION_DATE - execution_date to associate a run with
            spark.env.AIRFLOW_CTX_TASK_ID" - name of the Airflow Task to associate a run with
            spark.env.AIRFLOW_CTX_TRY_NUMBER" - try number of the Airflow Task to associate a run with

        Example:
            >>> get_dbnd_tracking_spark_conf()
            ['--conf', 'spark.env.AIRFLOW_CTX_DAG_ID={{dag.dag_id}}', '--conf', 'spark.env.AIRFLOW_CTX_EXECUTION_DATE={{ts}}', '--conf', 'spark.env.AIRFLOW_CTX_TASK_ID={{task.task_id}}', '--conf', 'spark.env.AIRFLOW_CTX_TRY_NUMBER={{task._try_number}}']

        Args:
            dag_id: name of the Airflow DAG to associate a run. By default set to {{dag.dag_id}} Airflow template
            task_id: name of the Airflow task to associate a run. By default set to {{task.task_id}} Airflow template
            execution_date: execution date as a string. By default set to {{ts}} Airflow template
            try_number: try number of Airflow task. By default set to {{task_instance._try_number}} Airflow template

        Returns:
            List of Airflow Ids as spark properties. Ready to concatenate to spark-submit command

    """

    conf_as_dict = get_dbnd_tracking_spark_conf_dict(
        dag_id, task_id, execution_date, try_number
    )
    conf = []
    for key in conf_as_dict.keys():
        conf.extend(["--conf", key + "=" + conf_as_dict[key]])
    return conf


def spark_submit_with_dbnd_tracking(command_as_list, dbnd_context=None):
    """ This functions augments spark-submit command with dbnd tracking metadata aka dbnd_context. If context
    is not provided, a default Airflow templates are set for DAG Id, Task Id, and execution date

    Adds 3 configuration properties to spark-submit to associate spark run with airflow DAG that initiated this run.
    These properties are
        spark.env.AIRFLOW_CTX_DAG_ID  -  name of the Airflow DAG to associate a run with
        spark.env.AIRFLOW_CTX_EXECUTION_DATE - execution_date to associate a run with
        spark.env.AIRFLOW_CTX_TASK_ID" - name of the Airflow Task to associate a run with

    Example:
        >>> spark_submit_with_dbnd_tracking(["spark-submit","my_script.py","my_param"])
        ['spark-submit', '--conf', 'spark.env.AIRFLOW_CTX_DAG_ID={{dag.dag_id}}', '--conf', 'spark.env.AIRFLOW_CTX_EXECUTION_DATE={{ts}}', '--conf', 'spark.env.AIRFLOW_CTX_TASK_ID={{task.task_id}}', 'my_script.py', 'my_param']

    Args:
        command_as_list: spark-submit command line as a list of strings.
        dbnd_context: optional dbnd_context provided by a user. I

    Returns:
        An updated spark-submit command including dbnd context.

    """

    if not dbnd_context:
        dbnd_context = get_dbnd_tracking_spark_conf()

    index = next(
        (
            command_as_list.index(elm)
            for elm in command_as_list
            if "spark-submit" in elm
        ),
        -1,
    )
    if index == -1:
        raise Exception("Failed to find spark-submit in %s" % " ".join(command_as_list))

    return command_as_list[0 : index + 1] + dbnd_context + command_as_list[index + 1 :]



from __future__ import print_function

import logging
import os

from dbnd._core.utils.basics.format_exception import format_exception_as_str
from dbnd_airflow.utils import dbnd_airflow_path


logger = logging.getLogger(__name__)


# avoid importing airflow on autocomplete (takes approximately 1s)


def setup_versioned_dags():
    # Check modules
    # Print relevant error
    from airflow.plugins_manager import plugins

    logging.info("Running setup for versioned dags!")
    from dbnd_airflow.airflow_override.dbnd_aiflow_webserver import (
        patch_airflow_create_app,
    )
    from dbnd_airflow.utils import dbnd_airflow_path
    from dbnd_airflow.plugins.dbnd_airflow_webserver_plugin import (
        DatabandAirflowWebserverPlugin,
    )

    for p in plugins:
        # the class instance can come from different location ( plugins folder for example)
        if DatabandAirflowWebserverPlugin.__name__ == p.__name__:
            return True

    command = "dbnd-airflow webserver"
    logger.info(
        "dbnd-airflow-versioned-dag is not installed. "
        "Please run 'pip install dbnd-airflow-versioned-dag' in order to run '{command}'.".format(
            command=command
        )
    )
    # we need it right now, plugins mechanism already scanned current folder
    patch_airflow_create_app()
    versioned_plugin_dir = dbnd_airflow_path(
        "plugins", "loadable_plugins", "patched_versioned_bag"
    )
    logger.info(
        "Linking dag versions plugins via AIRFLOW__CORE__PLUGINS_FOLDER to %s ",
        versioned_plugin_dir,
    )
    os.environ["AIRFLOW__CORE__PLUGINS_FOLDER"] = versioned_plugin_dir
    # ANOTHER OPTION WOULD BE TO ADD PERMAMENT LINK, however it will affect all other commands
    # plugin = dbnd_airflow_path(
    #     "plugins", "dbnd_airflow_webserver_plugin_linkable.py"
    # )
    # plugin_target = os.path.join(
    #     settings.PLUGINS_FOLDER, "dbnd_airflow_webserver_plugin.py"
    # )
    # link_dropin_file(
    #     plugin, plugin_target, unlink_first=False, name="web plugin"
    # )


def setup_scheduled_dags(sub_dir=None, unlink_first=True):
    from airflow import settings

    sub_dir = sub_dir or settings.DAGS_FOLDER

    if not sub_dir:
        raise Exception("Can't link scheduler: airflow's dag folder is undefined")

    source_path = dbnd_airflow_path("scheduler", "dags", "dbnd_dropin_scheduler.py")
    target_path = os.path.join(sub_dir, "dbnd_dropin_scheduler.py")

    if unlink_first and os.path.islink(target_path):
        try:
            logger.info("unlinking existing drop-in scheduler file at %s", target_path)
            os.unlink(target_path)
        except Exception:
            logger.error(
                "failed to unlink drop-in scheduler file at %s: %s",
                (target_path, format_exception_as_str()),
            )
            return

    if not os.path.exists(target_path):
        try:
            logger.info("Linking %s to %s.", source_path, target_path)
            if not os.path.exists(sub_dir):
                os.makedirs(sub_dir)
            os.symlink(source_path, target_path)
        except Exception:
            logger.error(
                "failed to link drop-in scheduler in the airflow dags_folder: %s"
                % format_exception_as_str()
            )



from airflow.models import TaskInstance
from airflow.utils.db import provide_session


@provide_session
def get_airflow_task_instance_state(task_run, session=None):
    TI = TaskInstance
    return (
        session.query(TI.state)
        .filter(
            TI.dag_id == task_run.task.ctrl.airflow_op.dag.dag_id,
            TI.task_id == task_run.task_af_id,
            TI.execution_date == task_run.run.execution_date,
        )
        .scalar()
    )


@provide_session
def get_airflow_task_instance(task_run, session=None):
    ti = (
        session.query(TaskInstance)
        .filter(
            TaskInstance.dag_id == task_run.task.ctrl.airflow_op.dag.dag_id,
            TaskInstance.task_id == task_run.task_af_id,
            TaskInstance.execution_date == task_run.run.execution_date,
        )
        .first()
    )
    return ti


@provide_session
def update_airflow_task_instance_in_db(ti, session=None):
    session.merge(ti)
    session.commit()



import logging

from dbnd_airflow.tracking.dbnd_dag_tracking import track_task


def _wrap_policy(policy):
    def new_policy(task):
        policy(task)
        track_task(task)

    return new_policy


def _patch_policy(module):
    new_policy = _wrap_policy(module.policy)
    module.policy = new_policy


def _add_tracking_to_policy():
    try:
        import airflow_local_settings

        _patch_policy(airflow_local_settings)
    except ImportError:
        pass

    from airflow import settings

    _patch_policy(settings)


def add_tracking_to_policy():
    """ Add tracking to all tasks as part of airflow policy """
    try:
        _add_tracking_to_policy()
    except Exception:
        logging.exception("Failed to add tracking in policy")


def patch_airflow_context_vars():
    """ used for tracking bash operators """
    import airflow
    from airflow.utils import operator_helpers
    from dbnd_airflow.airflow_override.operator_helpers import context_to_airflow_vars

    from dbnd._core.utils.object_utils import patch_models

    if hasattr(airflow.utils.operator_helpers, "context_to_airflow_vars"):
        patches = [
            (
                airflow.utils.operator_helpers,
                "context_to_airflow_vars",
                context_to_airflow_vars,
            )
        ]
        patch_models(patches)
